/*!
  * @license MIT (https://github.com/geotrev/rotom/blob/master/LICENSE)
  * Rotom v0.11.4 (https://github.com/geotrev/rotom#readme)
  * Copyright 2022 George Treviranus <geowtrev@gmail.com>
  */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('omdomdom')) :
  typeof define === 'function' && define.amd ? define(['exports', 'omdomdom'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.Rotom = {}, global.Omdomdom));
})(this, (function (exports, omdomdom) { 'use strict';

  const createUUID = () => {
    const base = Number.MAX_SAFE_INTEGER;
    return Math.floor(Math.random() * base).toString(36) + Math.abs(Date.now()).toString(36);
  };

  const getTypeTag = value => Object.prototype.toString.call(value).slice(8, -1).toLowerCase();
  const isPlainObject = value => getTypeTag(value) === "object";
  const isEmptyObject = value => {
    if (!isPlainObject(value)) return false;

    for (let key in value) {
      if (Object.prototype.hasOwnProperty.call(value, key)) {
        return false;
      }
    }

    return true;
  };
  const isFunction = value => getTypeTag(value) === "function";
  const isString = value => getTypeTag(value) === "string";
  const isUndefined = value => getTypeTag(value) === "undefined";

  const camelToKebab = value => value && value.match(/[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+/g).map(x => x.toLowerCase()).join("-");
  const kebabToCamel = value => value && value.split("-").map((word, i) => i > 0 ? word[0].toUpperCase() + word.slice(1).toLowerCase() : word.toLowerCase()).join("");

  const sanitizeString = value => {
    const dump = document.createElement("div");
    dump.textContent = value;
    return "" + dump.innerHTML;
  };

  function log(msg) {
    let type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "warn";
    console[type]("[RotomElement]: ".concat(msg));
  }

  function getTempName(name) {
    return "__".concat(name, "_initial");
  }

  const Attributes = {
    dir: "dir"
  };
  const AttributeValues = {
    ltr: "ltr"
  };
  const External = {
    rotomIdProperty: "rotomId",
    requestRender: "requestRender",
    render: "render",
    staticProperties: "properties",
    staticStyles: "styles",
    rotomIdAttribute: "rotom-id",
    onConnect: "onConnect",
    onMount: "onMount",
    onUpdate: "onUpdate",
    onPropertyChange: "onPropertyChange",
    onAttributeChange: "onAttributeChange",
    onUnmount: "onUnmount"
  };
  const Internal = {
    rotomId: Symbol("#rotomId"),
    vnode: Symbol("#vnode"),
    isFirstRender: Symbol("#isFirstRender"),
    propMap: Symbol("#propMap"),
    renderer: Symbol("#renderer"),
    initialize: Symbol("#initialize"),
    schedule: Symbol("#schedule"),
    runLifecycle: Symbol("#runLifecycle"),
    upgrade: Symbol("#upgrade"),
    upgradeProperties: Symbol("#upgradeProperties"),
    renderStyles: Symbol("#renderStyles"),
    renderDOM: Symbol("#renderDOM"),
    patch: Symbol("#patch"),
    destroy: Symbol("#destroy")
  };

  function setRenderer(renderer) {
    return {
      patch(element) {
        if (!window || !window.document) return;

        if (!isFunction(element[External.render])) {
          throw new Error("[RotomElement]: You must include a render method in element: '".concat(element.constructor.name, "'"));
        }

        renderer.patch(element);

        if (element[Internal.isFirstRender]) {
          element[Internal.isFirstRender] = false;
          element[Internal.runLifecycle](External.onMount);
        } else {
          element[Internal.runLifecycle](External.onUpdate);
        }
      },

      destroy(element) {
        if (!window || !window.document) return;
        element[Internal.isFirstRender] = true;
        element[Internal.runLifecycle](External.onUnmount);
        renderer.destroy(element);
        element[Internal.vnode] = null;
        const children = Array.apply(null, element.shadowRoot.childNodes);

        if (children.length) {
          Array.prototype.forEach.call(children, child => element.shadowRoot.removeChild(child));
        }
      }

    };
  }

  function createScheduler() {
    const requestAnimationFrame = window.requestAnimationFrame;
    const setTimeout = window.setTimeout;
    const FRAME_DURATION = 1000 / 60;
    let scheduled = null;

    function schedule(fn) {
      if (scheduled) return;

      if (isFunction(requestAnimationFrame)) {
        scheduled = requestAnimationFrame(() => {
          scheduled = null;
          fn();
        });
      } else if (isFunction(setTimeout)) {
        scheduled = setTimeout(() => {
          scheduled = null;
          fn();
        }, FRAME_DURATION);
      }
    }

    return fn => schedule(fn);
  }

  function validateType(Cls, propName, value, type) {
    if (isUndefined(type)) return;
    const evaluatedType = getTypeTag(value);
    if (evaluatedType === type) return;
    log("Property '".concat(propName, "' is invalid type of '").concat(evaluatedType, "'. Expected '").concat(type, "'. Check ").concat(Cls.constructor.name, "."));
  }

  const initializePropertyValue = (Cls, propName, _ref, privateName) => {
    let {
      default: defaultValue,
      type: propType,
      reflected = false,
      safe = false
    } = _ref;
    let initialValue;

    if (!isUndefined(Cls[propName])) {
      Cls[getTempName(propName)] = Cls[propName];
      initialValue = Cls[propName];
    } else if (isFunction(defaultValue)) {
      initialValue = defaultValue(Cls);
    } else {
      initialValue = defaultValue;
    }

    const valueEmpty = isUndefined(initialValue);
    const attrName = camelToKebab(propName);
    const attrValue = Cls.getAttribute(attrName);

    if (!valueEmpty) {
      {
        validateType(Cls, propName, initialValue, propType);
      }

      if (safe && isString(initialValue)) {
        initialValue = sanitizeString(initialValue);
      }

      Cls[privateName] = initialValue;
    } else if (reflected && attrValue) {
      initialValue = safe ? sanitizeString(attrValue) : attrValue;
      Cls[privateName] = initialValue;
    }

    if (reflected && !attrValue) {
      const initialAttrValue = initialValue ? String(initialValue) : "";
      Cls.setAttribute(attrName, initialAttrValue);
    }
  };

  const upgradeProperty = function (Cls, propName) {
    let configuration = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    if (Object.getOwnPropertyDescriptor(Object.getPrototypeOf(Cls), propName)) {
      return;
    }

    const privateName = Symbol(propName);
    const {
      type,
      reflected = false,
      safe = false
    } = configuration;

    if (reflected) {
      Cls[Internal.propMap][propName] = privateName;
    }

    initializePropertyValue(Cls, propName, configuration, privateName);
    Object.defineProperty(Cls, propName, {
      configurable: true,
      enumerable: true,

      get() {
        return Cls[privateName];
      },

      set(value) {
        if (value === Cls[privateName]) return;

        {
          validateType(Cls, propName, value, type);
        }

        if (reflected && !Cls[Internal.propMap[propName]]) {
          Cls[Internal.propMap][propName] = privateName;
        }

        const oldValue = Cls[privateName];

        if (!isUndefined(value)) {
          Cls[privateName] = safe && type === "string" && isString(value) && value !== "" ? sanitizeString(value) : value;
          Cls[Internal.runLifecycle](External.onPropertyChange, propName, oldValue, value);

          if (reflected) {
            const attribute = camelToKebab(propName);
            const attrValue = String(value);

            if (Cls.getAttribute(attribute) !== attrValue) {
              Cls.setAttribute(attribute, attrValue);
            }
          }
        } else {
          delete Cls[privateName];

          if (reflected) {
            delete Cls[Internal.propMap][propName];
          }

          const initialValue = Cls[getTempName(propName)];

          if (!isUndefined(initialValue)) {
            Cls[propName] = initialValue;
          }

          Cls[Internal.runLifecycle](External.onPropertyChange, propName, oldValue, value);

          if (reflected) {
            const attribute = camelToKebab(propName);
            Cls.removeAttribute(attribute);
          }
        }

        Cls[External.requestRender]();
      }

    });
  };

  const SHADOW_ROOT_MODE = "open";
  function rotomFactory(renderer) {
    return class RotomElement extends HTMLElement {
      constructor() {
        super();
        this[Internal.schedule] = createScheduler();
        this[Internal.renderer] = setRenderer(renderer);
        this.attachShadow({
          mode: SHADOW_ROOT_MODE
        });
        this[Internal.patch] = this[Internal.patch].bind(this);
        this[Internal.isFirstRender] = true;
        this[Internal.vnode] = null;
        this[Internal.rotomId] = createUUID();
        this[Internal.propMap] = {};
      }

      static get observedAttributes() {
        const properties = this[External.staticProperties];
        if (isEmptyObject(properties)) return [];
        let attributes = [];

        for (let propName in properties) {
          if (!properties[propName].reflected) continue;
          attributes.push(camelToKebab(propName));
        }

        return attributes;
      }

      adoptedCallback() {}

      attributeChangedCallback(name, oldValue, newValue) {
        if (oldValue !== newValue) {
          this[Internal.runLifecycle](External.onAttributeChange, name, oldValue, newValue);
          const propName = kebabToCamel(name);
          const privateName = this[Internal.propMap][propName];

          if (!isUndefined(privateName) && this[propName] !== newValue) {
            this[propName] = newValue;
          }
        }
      }

      connectedCallback() {
        if (!this.isConnected) return;
        this[Internal.upgrade]();
        this[Internal.runLifecycle](External.onConnect);
        this[Internal.renderStyles]();
        this[External.requestRender]();
      }

      disconnectedCallback() {
        this[Internal.destroy]();
      }

      get [External.rotomIdProperty]() {
        return this[Internal.rotomId];
      }

      [External.requestRender]() {
        this[Internal.schedule](this[Internal.patch]);
      }

      [Internal.runLifecycle](methodName) {
        if (isFunction(this[methodName])) {
          for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            args[_key - 1] = arguments[_key];
          }

          this[methodName](...args);
        }
      }

      [Internal.upgrade]() {
        this.setAttribute(External.rotomIdAttribute, this[External.rotomIdProperty]);
        this.setAttribute(Attributes.dir, String(document.dir || AttributeValues.ltr));
        this[Internal.upgradeProperties]();
      }

      [Internal.upgradeProperties]() {
        const properties = this.constructor[External.staticProperties];
        if (isEmptyObject(properties)) return;

        for (let propName in properties) {
          upgradeProperty(this, propName, properties[propName]);
        }
      }

      [Internal.destroy]() {
        this[Internal.renderer].destroy(this);
      }

      [Internal.patch]() {
        this[Internal.renderer].patch(this);
      }

      [Internal.renderStyles]() {
        const styles = this.constructor[External.staticStyles];
        if (!isString(styles)) return;
        const styleTag = document.createElement("style");
        styleTag.type = "text/css";
        styleTag.textContent = styles;
        this.shadowRoot.appendChild(styleTag);
      }

    };
  }

  const emptyVNode = {
    type: "comment",
    attributes: {},
    children: null,
    content: "",
    node: document.createComment("")
  };

  function getRenderState(element) {
    const domString = element[External.render]();

    if (!isString(domString)) {
      throw new Error("[RotomElement]: You attempted to render a non-string template in element: '".concat(element.constructor.name, "'."));
    }

    return domString;
  }

  function getInitialRenderState(element) {
    element[Internal.vnode] = omdomdom.create(getRenderState(element));
    omdomdom.render(element[Internal.vnode], element.shadowRoot);
  }

  function getNextRenderState(element) {
    let nextVnode = omdomdom.create(getRenderState(element));
    omdomdom.patch(nextVnode, element[Internal.vnode]);
    nextVnode = null;
  }

  const templateRenderer = {
    patch(element) {
      if (element[Internal.isFirstRender]) {
        getInitialRenderState(element);
      } else {
        getNextRenderState(element);
      }
    },

    destroy(element) {
      omdomdom.patch(emptyVNode, element[Internal.vnode]);
    }

  };

  function createDirectionObserver() {
    if (window.__ROTOM_ELEMENT__DIR_OBSERVER__) return;
    window.__ROTOM_ELEMENT__DIR_OBSERVER__ = true;

    const updateDirection = function () {
      let context = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : document;
      const nodes = Array.apply(null, context.querySelectorAll("[rotom-id]"));
      if (!nodes.length) return;
      nodes.forEach(node => {
        node.setAttribute("dir", String(document.dir || "ltr"));

        if (node.shadowRoot) {
          updateDirection(node.shadowRoot);
        }
      });
    };

    const mutationObserver = new MutationObserver(mutations => {
      mutations.forEach(mutation => {
        if (mutation.attributeName === "dir") {
          updateDirection();
        }
      });
    });
    mutationObserver.observe(document.documentElement, {
      attributes: true
    });
  }

  const register = (tag, RotomInstance) => {
    if (!customElements.get(tag)) {
      customElements.define(tag, RotomInstance);
    }
  };

  createDirectionObserver();
  const RotomElement = rotomFactory(templateRenderer);

  exports.RotomElement = RotomElement;
  exports.register = register;
  exports.validateType = validateType;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=rotom.template.js.map
